import java.awt.*;
import java.awt.image.*;

/**
 * Class ComplexRoot - Draws a plot of the behaviour of a polynomial
 * over a specified area when it is used with the Newton-Raphson method
 * of root finding. This usually produces a "fractal" type image.
 *
 * @author Matt Johnston
 */

public class ComplexPlotter {

private int plotwidth, plotheight;

private int canvaswidth = 200;
private int canvasheight = 200;

private double left = -5;
private double right = 5;
private double top = 5;
private double bottom = -5;

private int leftmargin;
private int rightmargin;
private int topmargin;
private int bottommargin;
private boolean drawscale = false;

// whether to produce a greyscale graph shaded by number of iterations.
private boolean greyscale = true;
// these are used for greyscale iteration shading
private int tonerange;
private int mintone;

private Canvas drawcanvas;
private Polynomial poly;
private ComplexRoot cplxroot;

private double scalefactorx, scalefactory;

// colour lookup array for different roots
private int[] colourlist;

private Complex[] roots;
private int numroots;

// lessprecision is used to determine maxprecision and colourprecision
private double lessprecision = 0.001; //this is for uniqueness testing
private double maxprecision; //this is for actually finding the roots
private double colourprecision; //this is for the colour lookup
private final static double PRECDIFF = 10; //specifies the difference between
											//the different precision values

// colouring variables
private static final int BLACK = 255 << 24;
private final float SAT = 0.82f;
private final float BRIGHT = 0.95f;

private static final int MAXROOTTRIALS = 1000;
private static final int MINCANVASSIZE = 50;

// label displacements
private static final int XAXISUP = -2;
private static final int XAXISRIGHT = -30;
private static final int YAXISUP = 10;
private static final int YAXISRIGHT = 1;

/**
 * Creates a new ComplexPlotter.
 */
public ComplexPlotter() {

}

/**
 * Returns a list of all the roots of a polynomial
 * @return A list of complex roots of the polynomial
 */
private Complex[] findRoots() {

	final int maxrootcount = poly.getDegree();
	Complex rootlist[] = new Complex[maxrootcount];
	Complex testroot;
	Complex[] returnroot;

	int rootcount = 0;
	double xrange = right - left;
	double yrange = top - bottom;

	// allow for repeated roots, give up after MAXROOTTRIALS
	for (int trial = 0; trial < MAXROOTTRIALS; trial++) {
		testroot = cplxroot.findRoot(new Complex(xrange*Math.random()+left,
												yrange*Math.random()+bottom));

		if (testroot == null) continue; // root cannot be found, we ignore

		if (uniqueroot(testroot, rootcount, rootlist)) {
			rootlist[rootcount] = new Complex(testroot);
			rootcount++;
			trial = 0; //keep going after finding roots, until MAXROOTTRIALS
		}

		if (rootcount == maxrootcount) break;

	}

	returnroot = new Complex[rootcount];
	for (int i = 0; i < rootcount; i++)
		returnroot[i] = rootlist[i];

	return returnroot;
}


/**
 * Tests if a root is a member of a list of previously found roots.
 * @param test The complex number to test for uniqueness.
 * @param index The number of elements in the list to check to. This allows
 * the list to be partially completed, with the remaining entries unspecified.
 * @param previous The list of preivously found complex roots.
 * @param precision The accuracy to compare roots at. This should be less
 * precise than the accuracy used to actually find the roots.
 */
private boolean uniqueroot(Complex test, int index, Complex[] previous) {

	for (int i = 0; i < index; i++) {
		if ((Math.abs(test.Re() - previous[i].Re()) < lessprecision)
			&& (Math.abs(test.Im() - previous[i].Im()) < lessprecision))
				return false;
	}

	return true;

}

/**
 * Sets the range of the area to draw a plot of.
 * @param left The leftmost extent of the plot. The default value is -5
 * @param right The rightmost extent of the plot. The default value is 5
 * @param bottom The bottommost extent of the plot. The default value is -5
 * @param top The topmost extent of the plot. The default value is 5
 */
public void setRange(double left, double right, double bottom, double top) {

	if (left > right) {
		System.out.println("Warning: left is greater than right,"
								+ " they have now been swapped");
		this.left = right;
		this.right = left;
	} else {
		this.left = left;
		this.right = right;
	}

	if (bottom > top) {
		System.out.println("Warning: bottom is greater than top,"
								+ " they have now been swapped");
		this.bottom = top;
		this.top = bottom;
	} else {
		this.bottom = bottom;
		this.top = top;
	}

}

/**
 * Sets the size of the canvas to use. If a scale is being shown, the
 * plot will not fill this entire area, as there will be margins.
 * @param width The width of the image to create.
 * This should be greater than 100. The default is 200
 * @param height The height of the image to create.
 * This should be greater than 100. The default is 200
 */
public void setCanvasSize(int width, int height) {

	if ((width < MINCANVASSIZE) || (height < MINCANVASSIZE)) {
		System.out.println("Warning: width and height must both be greater"
						+ " or equal to " + MINCANVASSIZE + ", they have now"
						+ " both been set to " + MINCANVASSIZE);

		canvaswidth = MINCANVASSIZE;
		canvasheight = MINCANVASSIZE;
		return;
	}

	canvaswidth = width;
	canvasheight = height;

}

/**
 * Sets the colouring method to use for the graph. This can either be
 * a different colour depending on the root converged to (the default),
 * or a greyscale shade depending on the number of iterations required.
 * @param greyscale Whether to use the greyscale method of colouring.
 */
public void setGreyScale(boolean greyscale) {

	this.greyscale = greyscale;

}

/**
 * Set whether to draw a scale beside the graph
 * @param drawscale True if you want to draw the scale. The default is false.
 */
public void setDrawScale(boolean drawscale) {

	this.drawscale = drawscale;

}

private void setupScale() {

	if (drawscale) {
		leftmargin = 25;
		rightmargin = 5;
		topmargin = 5;
		bottommargin = 15;
	} else {
		leftmargin = 0;
		rightmargin = 0;
		topmargin = 0;
		bottommargin = 0;
	}

	plotwidth = canvaswidth - leftmargin - rightmargin;
	plotheight = canvasheight - topmargin - bottommargin;

	scalefactorx = (this.right - this.left)/plotwidth;
	scalefactory = (this.top - this.bottom)/plotheight;

	// maxprecision is set depending on the polynomial's behaviour
	maxprecision = lessprecision / PRECDIFF; 

	colourprecision = lessprecision*PRECDIFF;


}

private void setupRootList() {

	roots = findRoots(); // a list hopefully containing all roots of poly

	numroots = roots.length;

	colourlist = new int[numroots];
	float hueoffset = (float)Math.random();

	for (int i = 0; i < numroots; i++) { // colours set by altering the hue
		Color tempcolour = new Color(java.awt.Color.HSBtoRGB(
								(i+1.0f)/numroots+hueoffset,SAT,BRIGHT));

		// colours are in Alpha, Red, Green, Blue format, one byte each.
		colourlist[i] = BLACK
							| tempcolour.getRed() << 16
							| tempcolour.getGreen() << 8
							| tempcolour.getBlue();
	}

}

private void plotColour() {

	int[] pix = new int[plotwidth];
	int index = 0;
	Complex testpoint = new Complex();

	for (int y = 0; y < plotheight; y++) {

		for (int x = 0; x < plotwidth; x++) {
			testpoint.setRe(getXPos(x));
			testpoint.setIm(getYPos(y));

			Complex currentpoint = cplxroot.findRoot(testpoint);
			pix[x] = getPlotColorInt(currentpoint);
		}

		Image img = drawcanvas.canvas.createImage(
				new MemoryImageSource(plotwidth,1,pix,0,plotwidth));

		drawcanvas.drawImage(img,leftmargin,y + topmargin);

	}

}

// setup greyscale colouring
private void setupTone() {

	int tempcount;
	int max = 0;
	int min = 255;
	double xrange = right - left;
	double yrange = top - bottom;

	//sample a range of points to determine representative iteration values
	for (int i = 0; i < 100; i++) {
		tempcount = cplxroot.findRootIterCount(new Complex(xrange*Math.random()+
											left,yrange*Math.random()+bottom));
		if (max < tempcount) max = tempcount;
		if (min > tempcount) min = tempcount;

	}

	mintone = min;
	tonerange = max - min;

}	

// retrieves a colour for the greyscale plotting, depending on the number
// of iterations required to converge.
private int getIterColour(int iterations) {

	int brightval;
	if (iterations <= mintone) {
		brightval = 255;
	} else {
		brightval = 255 - Math.min(
							(int)((iterations-mintone) * 255.0/tonerange),
							255);
	}

	return (BLACK) | (brightval << 16) | (brightval << 8) | brightval;

}

private void plotIterTone() {

	setupTone();

	int[] pix = new int[plotwidth];
	int index = 0;
	for (int y = 0; y < plotheight; y++) {
	for (int x = 0; x < plotwidth; x++) 
		pix[x] = getIterColour(cplxroot.findRootIterCount(
										new Complex(getXPos(x), getYPos(y))));

		Image img = drawcanvas.canvas.createImage(
				new MemoryImageSource(plotwidth,1,pix,0,plotwidth));

		drawcanvas.drawImage(img,leftmargin,y + topmargin);
	}

}


/**
 * Draw a graphical representation of the behaviour of the polynomial
 * function with the Newton Raphson method. By default, each point will
 * be coloured depending on the root it is converged on. Alternately,
 * if setGreyScale has been set, then each point will be coloured a
 * shade of grey depending on the number of iterations required to reach
 * zero. This mode of drawing has quite different behaviour to the default.
 * @param pol The polynomial function to use.
 */
public void drawRoots(Polynomial pol) {

	drawRoots(pol, null, null);

}

public boolean drawRoots(Polynomial pol, String filename, String format) {

	if (pol.getDegree() == 0) {
		System.out.println("No roots for this function");
		return false;
	}

	this.poly = pol;

	setupScale();

	cplxroot = new ComplexRoot(poly, maxprecision);
	setupRootList(); //create a list of all roots of the Polynomial

	drawcanvas = new Canvas("Complex Roots",canvaswidth,canvasheight);
	drawcanvas.setVisible(true);

	if (drawscale) drawScale();

	if (!greyscale)
		plotColour();
	else
		plotIterTone();

	if (filename != null) {
		System.out.println("Starting writing file");
		boolean ret = drawcanvas.writeFile(filename, format);
		System.out.println("Done writing file");
		return ret;
	}
	return true;
}

// used to draw text at a particular point, handles colour changing
private void drawText(double value, int x, int y) {

	java.text.DecimalFormat df;

	// numbers between 0.001 and 1000 are normal, others are exp. notation
	if ((Math.abs(value) < 1000) && (Math.abs(value) > 0.001) || (value==0))
		df = new java.text.DecimalFormat("0.##");
	else 
		df = new java.text.DecimalFormat("#0.#E0");

	df.setDecimalSeparatorAlwaysShown(false);

	drawcanvas.drawString(df.format(value), x, y);

}


private void drawScale() {

	// x axis
	drawText(left, leftmargin, canvasheight + XAXISUP);
	drawText(right, canvaswidth + XAXISRIGHT, canvasheight + XAXISUP);
	drawText((right-left)/2, leftmargin + (plotwidth + XAXISRIGHT)/2, 
							canvasheight + XAXISUP);

	// y axis
	drawText(top, YAXISRIGHT, topmargin + YAXISUP);
	drawText(bottom, YAXISRIGHT, topmargin + plotheight);
	drawText((top-bottom)/2, YAXISRIGHT, topmargin + (plotheight / 2));

}

private double getXPos(double x) {

	return left + x * scalefactorx;
}

private double getYPos(double y) {

	return top - y * scalefactory;

}

private int getPlotColorInt(Complex point) {

	if (point == null) return BLACK; 

	for (int i = 0; i < numroots; i++)
		if ((Math.abs(point.Re() - roots[i].Re()) < colourprecision)
			&& (Math.abs(point.Im() - roots[i].Im()) < colourprecision))
				return colourlist[i];

	// this happens only if root is not in the list, and shouldn't actually
	// be reached during normal operation.
	return BLACK;

}

}

